from ..uftp import DsoPortfolioQuery, DsoPortfolioUpdate, UsefRole
from .base_client import ShapeshifterClient


class ShapeshifterDsoCroClient(ShapeshifterClient):
    """
    Client that allows the DSO to connect to the CRO.
    """

    sender_role = UsefRole.DSO
    recipient_role = UsefRole.CRO

    def send_dso_portfolio_update(self, message: DsoPortfolioUpdate) -> None:
        """
        The DSOPortfolioUpdate is used by the DSO to indicate on which
        congestion points it wants to engage in flexibility trading.
        """
        self._send_message(message)

    def send_dso_portfolio_query(self, message: DsoPortfolioQuery) -> None:
        """
        DSOPortfolioQuery is used by DSOs to discover which AGRs represent
        connections on its registered congestion point(s).
        """
        self._send_message(message)



from dataclasses import dataclass

from .payload_message import PayloadMessage, PayloadMessageResponse


@dataclass(kw_only=True)
class TestMessage(PayloadMessage):
    __test__ = False  # Tell pytest to ignore this class


@dataclass(kw_only=True)
class TestMessageResponse(PayloadMessageResponse):
    __test__ = False  # Tell pytest to ignore this class



import re
from concurrent.futures import ThreadPoolExecutor
from threading import Thread
from time import sleep

import uvicorn
from fastapi import FastAPI, Response
from fastapi.exceptions import HTTPException
from fastapi_xml import XmlAppResponse, XmlRoute

from .. import transport
from ..client import client_map
from ..exceptions import (
    FunctionalException,
    InvalidMessageException,
    InvalidSenderException,
    TransportException,
)
from ..logging import logger
from ..uftp import (
    AcceptedRejected,
    PayloadMessage,
    SignedMessage,
    TestMessage,
    TestMessageResponse,
    UsefRole,
    request_response_map,
)


class ShapeshifterService():
    """
    Basis for all Shapeshifter Services. Defines the web service, the
    message ingestion, the post-processing queue mechanics, and
    threading and context options.
    """

    sender_domain = None
    sender_role = None
    acceptable_messages = []

    num_inbound_threads = 10
    num_outbound_threads = 10

    def __init__(
        self,
        sender_domain,
        signing_key,
        key_lookup_function=None,
        endpoint_lookup_function=None,
        oauth_lookup_function=None,
        host: str = "0.0.0.0",
        port: int = 8080,
        path: str = "/shapeshifter/api/v3/message",
        version: str = "3.1.0"
    ):
        """
        :param sender_domain: our sender domain (FQDN) that the recipient uses to look us up.
        :param signing_key: the private singing key that we use to sign outgoing messages.
        :param key_lookup_function: A callable that takes a (sender_domain, sender_role)
                                  pair and returns a verify_key (str or bytes).
                                  Omit parameter to use DNS for key lookup.
        :param key_lookup_function: A callable that takes a (sender_domain, sender_role)
                                  pair and returns a full endpoint URL (str).
                                  Omit parameter to use DNS for endpoint lookup.
        :param oauth_lookup_function: A callable that takes a (sender_domain, sender_role)
                                      pair and returns in instance of shapeshifter_uftp.OAuthClient
                                      if OAuth authentication is required.
        :param host: the host to bind the server to (usually 127.0.0.1 or 0.0.0.0)
        :param port: the port to bind the server to (default: 8080)
        :param path: the URL path that the server listens on (default: /shapeshifter/api/v3/message)
        """

        if version not in ("3.0.0", "3.1.0"):
            raise ValueError(f"'version' must be one of '3.0.0' or '3.1.0', not {version}")

        self.version = version

        # Set the sender domain, which is used
        # to identify us to the other party.
        self.sender_domain = sender_domain

        # The signing key is used to sign outgoing messages. The
        # corresponding public key should be published via DNS or
        # given to the recipient out-of-band.
        self.signing_key = signing_key

        # The key lookup method is used to look up keys for the other
        # party. If omitted, use DNS lookups using well-known DNS names.
        self.key_lookup_function = key_lookup_function or transport.get_key

        # The endpoint lookup method is used to look up the endpoint
        # that we send all messages to. If omitted, use DNS lookups
        # using well-known DNS names.
        self.endpoint_lookup_function = endpoint_lookup_function or transport.get_endpoint

        # The OAuth lookup function is used to get the OAuth instance
        # used to authenticate outgoing requests.
        self.oauth_lookup_function = oauth_lookup_function

        # The FastAPI web app takes care of routing messages to the
        # (one) endpoint, and by virtue of FastAPI-XML convert the
        # python-friendly objects into XML and vice versa.
        self.app = FastAPI(default_response_class=XmlAppResponse)
        self.app.router.route_class = XmlRoute
        self.app.router.add_api_route(
            path,
            endpoint=self._receive_message,
            response_model=None,
            methods=["POST"],
            status_code=200,
        )

        # The web server hosts the FastAPI application and takes care
        # of the HTTP transport.
        config = uvicorn.Config(app=self.app, host=host, port=port)
        self.server = uvicorn.Server(config)
        self.server_thread = None

        # Create an inbound executor worker
        self.inbound_executor = ThreadPoolExecutor(max_workers=self.num_inbound_threads)
        self.outbound_executor = ThreadPoolExecutor(max_workers=self.num_outbound_threads)


    def run(self):
        """
        Start the web server that hosts the FastAPI application. Other
        participants can now send messages to us.
        """
        # Start the service and start accepting incoming requests.
        self.server.run()

    def run_in_thread(self):
        """
        Run the service in a background thread.
        """
        self.server_thread = Thread(target=self.run)
        self.server_thread.start()
        while not self.server.started:
            sleep(0.1)

    def stop(self):
        """
        Stop the service if it was running in a separate thread.
        """
        self.server.should_exit = True
        if self.server_thread and self.server_thread.is_alive():
            self.server_thread.join()
        self.server_thread = None


    # ------------------------------------------------------------ #
    #   Message handling and processing methods, internal to the   #
    #              Shapeshifter UFTP implementation.               #
    # ------------------------------------------------------------ #

    def _receive_message(self, message: SignedMessage) -> Response:
        """
        The default entrypoint for the route. This will unpack the
        message and validate the signature. It will thes pass the
        PayloadMessago to the pre processing function for a
        response.
        """
        logger.info(f"Got a request: {message}")
        # Get the public key that is used to decrypt the message
        signing_key = self.key_lookup_function(
            message.sender_domain, message.sender_role.value
        )

        logger.debug(f"The signing key is {signing_key}")

        # Unseal the message, returning an error if required
        try:
            unsealed_message = transport.unseal_message(message.body, signing_key)

            # Verify that the sender_domain inside the message is the
            # same as the sender_domain of the SignedMessage
            # envelope
            if unsealed_message.sender_domain != message.sender_domain:
                logger.warning(
                    "Received a message with mismatching sender_domain in the "
                    f"SignedMessage envelope ({message.sender_domain}) and the "
                    f"inner PayloadMessage ({unsealed_message.sender_domain})"
                )
                raise InvalidSenderException()

            if type(unsealed_message) not in self.acceptable_messages:
                logger.warning(
                    f"Received a misdirected message of type {unsealed_message.__class__.__name__} "
                    f"from {unsealed_message.sender_domain}.")
                raise InvalidMessageException(unsealed_message)

        except TransportException as err:
            logger.warning(f"The original transport error is {err.__class__.__name__}: {err}")
            raise HTTPException(err.http_status_code) from err

        except FunctionalException as err:
            self.outbound_executor.submit(self._reject_message, message, unsealed_message, err.rejection_reason)

        else:
            # If the initial checks passed, process the message in the
            # user-defined pipeline.
            self.inbound_executor.submit(self._process_message, unsealed_message, message.sender_role)

        return Response(status_code=200)

    def _process_message(self, message: PayloadMessage, sender_role: UsefRole):
        """
        Find the relevant post-processing method to handle the message
        outside of the request context, and run it.
        """
        process_method_name = f"process_{snake_case(message.__class__.__name__)}"
        process_method = getattr(self, process_method_name)
        try:
            if isinstance(message, TestMessage):
                process_method(message, sender_role)
            else:
                process_method(message)
        except Exception as err:  # pylint: disable=broad-exception-caught
            logger.error(
                f"An error occurred during the post-processing of a {message.__class__.__name__} message."
                f"{err.__class__.__name__}: {err}"
            )

    def _get_client(self, recipient_domain: str, recipient_role: UsefRole, version: str = "3.1.0"):
        """
        Method to get a relevant client to communicate to the
        indicated participant.
        """
        client_cls = client_map[(self.sender_role, recipient_role)]
        recipient_endpoint = self.endpoint_lookup_function(recipient_domain, recipient_role)
        recipient_signing_key = self.key_lookup_function(recipient_domain, recipient_role)
        oauth_client = self.oauth_lookup_function(recipient_domain, recipient_role) if self.oauth_lookup_function else None
        return client_cls(
            sender_domain = self.sender_domain,
            signing_key = self.signing_key,
            recipient_domain = recipient_domain,
            recipient_endpoint = recipient_endpoint,
            recipient_signing_key = recipient_signing_key,
            oauth_client = oauth_client,
            version=version,
        )

    def _reject_message(self, message, unsealed_message, reason):
        """
        Send a rejection to the sending party.
        """
        if type(unsealed_message) not in request_response_map:
            return

        client = self._get_client(message.sender_domain, message.sender_role, unsealed_message.version)
        response_type = request_response_map[type(unsealed_message)]
        response_id_field = snake_case(type(unsealed_message).__name__) + "_message_id"
        message_contents = {
            "recipient_domain": message.sender_domain,
            "conversation_id": unsealed_message.conversation_id,
            "result": AcceptedRejected.REJECTED,
            "rejection_reason": reason,
            response_id_field: unsealed_message.message_id
        }
        response_message = response_type(**message_contents)
        client._send_message(response_message)

    def __enter__(self):
        """
        Context-manager method that allows an instance of this class to be
        used in a temporary context.

        Starts the uvicorn server in a separate thread and waits for it
        to be started. Useful in testing scenarios.

        Usage:

        with MyShapeshifterServiceSubClass(...) as server:
            # your test code here, server exists cleanly when leaving
            # the 'with' block.
        """
        self.run_in_thread()
        return self

    def __exit__(self, *args, **kwargs):
        """
        Tell uvicorn we should exit and wait for the thread to finish.
        """
        self.stop()

    # ------------------------------------------------------------ #
    #                Common messages to all parties                #
    # ------------------------------------------------------------ #

    def process_test_message(self, message: TestMessage, sender_role: UsefRole):
        logger.info(
            "Received a TestMessage, will respond with TestMessageResponse. "
            f"Implement the process_test_message method on your {self.__class__.__qualname__} "
            f"to implement custom behavior. The message was: {message}"
        )
        response = TestMessageResponse(conversation_id=message.conversation_id)
        client = self._get_client(message.sender_domain, sender_role, version=message.version or self.version)
        client.send_test_message_response(response)

    def process_test_message_response(self, message: TestMessage):
        logger.info(f"Received a TestMessageResponse: {message}")



def snake_case(text):
    """
    Convert text from CamelCase to snake_case.
    """
    return re.sub(r"(.)([A-Z][a-z])", r"\1_\2", text).lower()



from dataclasses import dataclass, field
from typing import List, Optional

from xsdata.models.datatype import XmlDate

from ..defaults import DEFAULT_TIME_ZONE
from ..enums import RedispatchBy
from ..validations import validate_list
from .payload_message import PayloadMessage, PayloadMessageResponse


@dataclass(kw_only=True)
class AgrPortfolioQueryResponseConnection:
    """
    :ivar entity_address: EntityAddress of the Connection.
    """
    class Meta:
        name = "AGRPortfolioQueryResponseConnection"

    entity_address: str = field(
        metadata={
            "name": "EntityAddress",
            "type": "Attribute",
            "required": True,
            "pattern": r"(ea1\.[0-9]{4}-[0-9]{2}\..{1,244}:.{1,244}|ean\.[0-9]{12,34})",
        }
    )

@dataclass(kw_only=True)
class AgrPortfolioQueryResponseCongestionPoint:
    """
    :ivar connection:
    :ivar entity_address: EntityAddress of the CongestionPoint.
    :ivar mutex_offers_supported: Indicates whether the DSO accepts
        mutual exclusive FlexOffers on this CongestionPoint.
    :ivar day_ahead_redispatch_by: Indicates which party is responsible
        for day-ahead redispatch.
    :ivar intraday_redispatch_by: Indicates which party is responsible
        for intraday ahead redispatch, AGR or DSO. If not specified,
        there will be no intraday trading on this CongestionPoint.
    """
    class Meta:
        name = "AGRPortfolioQueryResponseCongestionPoint"

    connections: List[AgrPortfolioQueryResponseConnection] = field(
        default_factory=list,
        metadata={
            "name": "Connection",
            "type": "Element",
            "min_occurs": 1,
        }
    )
    entity_address: str = field(
        metadata={
            "name": "EntityAddress",
            "type": "Attribute",
            "required": True,
            "pattern": r"(ea1\.[0-9]{4}-[0-9]{2}\..{1,244}:.{1,244}|ean\.[0-9]{12,34})",
        }
    )
    mutex_offers_supported: bool = field(
        metadata={
            "name": "MutexOffersSupported",
            "type": "Attribute",
            "required": True,
        }
    )
    day_ahead_redispatch_by: RedispatchBy = field(
        metadata={
            "name": "DayAheadRedispatchBy",
            "type": "Attribute",
            "required": True,
        }
    )
    intraday_redispatch_by: Optional[RedispatchBy] = field(
        default=None,
        metadata={
            "name": "IntradayRedispatchBy",
            "type": "Attribute",
        }
    )

    def __post_init__(self):
        validate_list('connections', self.connections, AgrPortfolioQueryResponseConnection, 1)





@dataclass(kw_only=True)
class AgrPortfolioQueryResponseDSOPortfolio:
    class Meta:
        name = "AGRPortfolioQueryResponseDSOPortfolio"

    congestion_points: List[AgrPortfolioQueryResponseCongestionPoint] = field(
        default_factory=list,
        metadata={
            "name": "CongestionPoint",
            "type": "Element",
            "min_occurs": 1,
        }
    )
    dso_domain: str = field(
        metadata={
            "name": "DSO-Domain",
            "type": "Attribute",
            "required": True,
            "pattern": r"([a-z0-9]+(-[a-z0-9]+)*\.)+[a-z]{2,}",
        }
    )

    def __post_init__(self):
        validate_list('congestion_points', self.congestion_points, AgrPortfolioQueryResponseCongestionPoint, 1)


@dataclass(kw_only=True)
class AgrPortfolioQueryResponseDSOView:
    class Meta:
        name = "AGRPortfolioQueryResponseDSOView"

    dso_portfolios: List[AgrPortfolioQueryResponseDSOPortfolio] = field(
        default_factory=list,
        metadata={
            "name": "DSO-Portfolio",
            "type": "Element",
            "min_occurs": 1,
        }
    )
    connections: List[AgrPortfolioQueryResponseConnection] = field(
        default_factory=list,
        metadata={
            "name": "Connection",
            "type": "Element",
        }
    )

    def __post_init__(self):
        validate_list('dso_portfolios', self.dso_portfolios, AgrPortfolioQueryResponseDSOPortfolio, 1)




@dataclass(kw_only=True)
class AgrPortfolioQueryResponse(PayloadMessageResponse):
    """
    :ivar dso_view:
    :ivar time_zone: Time zone ID (as per the IANA time zone database,
        http://www.iana.org/time-zones, for example: Europe/Amsterdam)
        indicating the UTC offset that applies to the Period referenced
        in this message. Although the time zone is a market-wide fixed
        value, making this assumption explicit in each message is
        important for validation purposes, allowing implementations to
        reject messages with an errant UTC offset.
    :ivar period: The Period that the portfolio is valid.
    """
    class Meta:
        name = "AGRPortfolioQueryResponse"

    agr_portfolio_query_message_id: str = field(
        metadata={
            "name": "AGRPortfolioQueryMessageID",
            "type": "Attribute",
            "required": True,
            "pattern": r"[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-[0-9A-Fa-f]{4}-[0-9A-Fa-f]{4}-[0-9A-Fa-f]{12}",
        }
    )

    dso_views: List[AgrPortfolioQueryResponseDSOView] = field(
        default_factory=list,
        metadata={
            "name": "DSO-View",
            "type": "Element",
            "min_occurs": 1,
        }
    )
    time_zone: str = field(
        default=DEFAULT_TIME_ZONE,
        metadata={
            "name": "TimeZone",
            "type": "Attribute",
            "required": True,
            "pattern": r"(Africa|America|Australia|Europe|Pacific)/[a-zA-Z0-9_/]{3,}",
        }
    )
    period: XmlDate = field(
        metadata={
            "name": "Period",
            "type": "Attribute",
            "format": "%Y-%m-%d",
            "required": True,
        }
    )

    def __post_init__(self):
        self.dso_views = validate_list('dso_views', self.dso_views, AgrPortfolioQueryResponseDSOView, 1)


@dataclass(kw_only=True)
class AgrPortfolioQuery(PayloadMessage):
    """
    :ivar time_zone: Time zone ID (as per the IANA time zone database,
        http://www.iana.org/time-zones, for example: Europe/Amsterdam)
        indicating the UTC offset that applies to the Period referenced
        in this message. Although the time zone is a market-wide fixed
        value, making this assumption explicit in each message is
        important for validation purposes, allowing implementations to
        reject messages with an errant UTC offset.
    :ivar period: The Period for which the AGR requests the portfolio
        information.
    """

    class Meta:
        name = "AGRPortfolioQuery"

    time_zone: str = field(
        default=DEFAULT_TIME_ZONE,
        metadata={
            "name": "TimeZone",
            "type": "Attribute",
            "required": True,
            "pattern": r"(Africa|America|Australia|Europe|Pacific)/[a-zA-Z0-9_/]{3,}",
        }
    )
    period: XmlDate = field(
        metadata={
            "name": "Period",
            "type": "Attribute",
            "format": "%Y-%m-%d",
            "required": True,
        }
    )
